package org.gemoc.execution.sequential.javaengine.ui.debug;

import com.google.common.base.Objects;
import fr.inria.diverse.trace.api.IStep;
import fr.inria.diverse.trace.api.IValueTrace;
import fr.inria.diverse.trace.gemoc.api.IGemocTraceManager;
import fr.inria.diverse.trace.gemoc.api.IMultiDimensionalTraceAddon;
import fr.obeo.dsl.debug.ide.event.IDSLDebugEventProcessor;
import java.util.ArrayList;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.BiPredicate;
import java.util.function.Consumer;
import java.util.function.Supplier;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EOperation;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.xtext.naming.DefaultDeclarativeQualifiedNameProvider;
import org.eclipse.xtext.naming.QualifiedName;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.gemoc.execution.sequential.javaengine.ui.Activator;
import org.gemoc.execution.sequential.javaengine.ui.debug.GenericSequentialModelDebugger;
import org.gemoc.execution.sequential.javaengine.ui.debug.WrapperOmniscientDebugTimeLine;
import org.gemoc.executionframework.engine.core.AbstractSequentialExecutionEngine;
import org.gemoc.executionframework.engine.mse.MSE;
import org.gemoc.executionframework.engine.mse.MSEOccurrence;
import org.gemoc.xdsmlframework.api.core.IBasicExecutionEngine;
import org.gemoc.xdsmlframework.api.core.ISequentialExecutionEngine;

@SuppressWarnings("all")
public class OmniscientGenericSequentialModelDebugger extends GenericSequentialModelDebugger {
  private final IMultiDimensionalTraceAddon traceAddon;
  
  /**
   * -1 means we are on the last recorded state.
   * Otherwise value of the last jump index.
   */
  private int lastJumpIndex;
  
  private boolean inThePast = false;
  
  private final List<IStep.StepEvent> stepEvents = new ArrayList<IStep.StepEvent>();
  
  private int currentEvent = (-1);
  
  private int steppingOverStackFrameIndex = (-1);
  
  private int steppingReturnStackFrameIndex = (-1);
  
  public OmniscientGenericSequentialModelDebugger(final IDSLDebugEventProcessor target, final ISequentialExecutionEngine engine, final IMultiDimensionalTraceAddon addon) {
    super(target, engine);
    this.traceAddon = addon;
    WrapperOmniscientDebugTimeLine _wrapperOmniscientDebugTimeLine = new WrapperOmniscientDebugTimeLine(this);
    this.traceAddon.setTimeLineNotifier(_wrapperOmniscientDebugTimeLine);
    this.lastJumpIndex = (-1);
  }
  
  public boolean isInReplayMode() {
    return this.inThePast;
  }
  
  private MSE getMSEFromStep(final EObject caller, final IStep step) {
    MSE mse = null;
    if ((caller instanceof MSEOccurrence)) {
      MSE _mse = ((MSEOccurrence) caller).getMse();
      mse = _mse;
    } else {
      String _containingClassName = step.getContainingClassName();
      String _operationName = step.getOperationName();
      MSE _findOrCreateMSE = ((AbstractSequentialExecutionEngine) this.engine).findOrCreateMSE(caller, _containingClassName, _operationName);
      mse = _findOrCreateMSE;
    }
    return mse;
  }
  
  private void pushStackFrame(final String threadName, final IStep step) {
    EObject caller = null;
    MSE mse = null;
    String name = null;
    Map<String, Object> _parameters = step.getParameters();
    Set<Map.Entry<String, Object>> _entrySet = _parameters.entrySet();
    final Function1<Map.Entry<String, Object>, Boolean> _function = new Function1<Map.Entry<String, Object>, Boolean>() {
      @Override
      public Boolean apply(final Map.Entry<String, Object> es) {
        String _key = es.getKey();
        return Boolean.valueOf(_key.equals("caller"));
      }
    };
    final Map.Entry<String, Object> callerEntry = IterableExtensions.<Map.Entry<String, Object>>findFirst(_entrySet, _function);
    boolean _notEquals = (!Objects.equal(callerEntry, null));
    if (_notEquals) {
      Object _value = callerEntry.getValue();
      final EObject entryValue = ((EObject) _value);
      MSE _mSEFromStep = this.getMSEFromStep(entryValue, step);
      mse = _mSEFromStep;
      String _name = mse.getName();
      name = _name;
    } else {
      IStep _parentStep = step.getParentStep();
      Map<String, Object> _parameters_1 = _parentStep.getParameters();
      Object _get = _parameters_1.get("this");
      final EObject parentStep = ((EObject) _get);
      MSE _mSEFromStep_1 = this.getMSEFromStep(parentStep, step);
      mse = _mSEFromStep_1;
      String _name_1 = mse.getName();
      String _plus = (_name_1 + "_implicitStep");
      name = _plus;
    }
    EObject _caller = mse.getCaller();
    caller = _caller;
    EClass _eClass = caller.eClass();
    String _name_2 = _eClass.getName();
    String _plus_1 = (_name_2 + " (");
    String _plus_2 = (_plus_1 + name);
    String _plus_3 = (_plus_2 + ") [");
    String _string = caller.toString();
    String _plus_4 = (_plus_3 + _string);
    String _plus_5 = (_plus_4 + "]");
    name = _plus_5;
    final DefaultDeclarativeQualifiedNameProvider nameprovider = new DefaultDeclarativeQualifiedNameProvider();
    final QualifiedName qname = nameprovider.getFullyQualifiedName(caller);
    String _xifexpression = null;
    if ((qname != null)) {
      _xifexpression = qname.toString();
    } else {
      _xifexpression = caller.toString();
    }
    final String objectName = _xifexpression;
    EOperation _action = mse.getAction();
    String _name_3 = null;
    if (_action!=null) {
      _name_3=_action.getName();
    }
    final String opName = _name_3;
    EClass _eClass_1 = caller.eClass();
    final String callerType = _eClass_1.getName();
    final String prettyName = (((((("(" + callerType) + ") ") + objectName) + " -> ") + opName) + "()");
    this.pushStackFrame(threadName, prettyName, caller, caller);
  }
  
  private void updateStateEvents(final int state) {
    this.stepEvents.clear();
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    List<IStep.StepEvent> _eventsForState = _traceManager.getEventsForState(state);
    this.stepEvents.addAll(_eventsForState);
    this.currentEvent = 0;
  }
  
  private void loadLastState(final String threadName) {
    if ((this.lastJumpIndex != (-1))) {
      int _lastIndex = this.getLastIndex();
      this.jumpToState(_lastIndex);
    }
    final LinkedList<IStep> virtualStack = new LinkedList<IStep>();
    final int size = this.stepEvents.size();
    while ((this.currentEvent < size)) {
      {
        final IStep.StepEvent event = this.stepEvents.get(this.currentEvent);
        if (event.start) {
          virtualStack.addLast(event.step);
        } else {
          boolean _isEmpty = virtualStack.isEmpty();
          if (_isEmpty) {
            this.popStackFrame(threadName);
          } else {
            virtualStack.removeFirst();
          }
        }
        this.currentEvent++;
      }
    }
    final Consumer<IStep> _function = new Consumer<IStep>() {
      @Override
      public void accept(final IStep s) {
        OmniscientGenericSequentialModelDebugger.this.pushStackFrame(threadName, s);
      }
    };
    virtualStack.forEach(_function);
    this.inThePast = false;
    this.currentEvent = (-1);
  }
  
  private void findPreviousStartedStep(final String threadName) {
    int itr = this.currentEvent;
    if ((itr > 0)) {
      itr--;
      IStep.StepEvent event = this.stepEvents.get(itr);
      while (((!event.start) && (itr > 0))) {
        {
          itr--;
          IStep.StepEvent _get = this.stepEvents.get(itr);
          event = _get;
        }
      }
      if (event.start) {
        Map<String, Object> _parameters = event.step.getParameters();
        final Object step = _parameters.get("this");
        int _currentStateIndex = this.getCurrentStateIndex();
        this.jumpToState(_currentStateIndex);
        this.findCorrespondingStepEvent(threadName, step);
      } else {
        if ((this.lastJumpIndex != 0)) {
          IStep.StepEvent _get = this.stepEvents.get(0);
          Map<String, Object> _parameters_1 = _get.step.getParameters();
          final Object step_1 = _parameters_1.get("this");
          int _currentStateIndex_1 = this.getCurrentStateIndex();
          int _minus = (_currentStateIndex_1 - 1);
          this.jumpToState(_minus);
          this.findCorrespondingStepEvent(threadName, step_1);
        } else {
        }
      }
    } else {
      if ((this.lastJumpIndex != 0)) {
        IStep.StepEvent _get_1 = this.stepEvents.get(0);
        Map<String, Object> _parameters_2 = _get_1.step.getParameters();
        final Object step_2 = _parameters_2.get("this");
        int _currentStateIndex_2 = this.getCurrentStateIndex();
        int _minus_1 = (_currentStateIndex_2 - 1);
        this.jumpToState(_minus_1);
        this.findCorrespondingStepEvent(threadName, step_2);
      } else {
      }
    }
  }
  
  private void findNextStartedStep(final String threadName) {
    final int size = this.stepEvents.size();
    if ((this.currentEvent < size)) {
      IStep.StepEvent event = this.stepEvents.get(this.currentEvent);
      while (((!event.start) && (this.currentEvent < (size - 1)))) {
        {
          this.popStackFrame(threadName);
          this.currentEvent++;
          IStep.StepEvent _get = this.stepEvents.get(this.currentEvent);
          event = _get;
        }
      }
      if (event.start) {
        this.pushStackFrame(threadName, event.step);
      } else {
        if ((this.lastJumpIndex == (-1))) {
          this.popStackFrame(threadName);
        } else {
          throw new IllegalStateException("State events ended before a start event was found");
        }
      }
    }
  }
  
  private IStep findPreviousEndedStep(final String threadName) {
    int itr = this.currentEvent;
    if ((itr > 0)) {
      itr--;
      IStep.StepEvent event = this.stepEvents.get(itr);
      while ((event.start && (itr > 0))) {
        {
          itr--;
          IStep.StepEvent _get = this.stepEvents.get(itr);
          event = _get;
        }
      }
      if ((!event.start)) {
        return event.step;
      } else {
        if ((this.lastJumpIndex != 0)) {
          throw new IllegalStateException("State events ended before an end event was found");
        } else {
          return null;
        }
      }
    } else {
      throw new IllegalStateException("State events ended before an end event was found");
    }
  }
  
  private void findCorrespondingStepEvent(final String threadName, final Object step) {
    final int size = this.stepEvents.size();
    if ((this.currentEvent < size)) {
      IStep.StepEvent event = this.stepEvents.get(this.currentEvent);
      Map<String, Object> _parameters = event.step.getParameters();
      Object obj = _parameters.get("this");
      LinkedList<IStep> virtualStack = new LinkedList<IStep>();
      while (((!Objects.equal(obj, step)) && (this.currentEvent < (size - 1)))) {
        {
          if (event.start) {
            virtualStack.push(event.step);
          } else {
            boolean _isEmpty = virtualStack.isEmpty();
            if (_isEmpty) {
              this.popStackFrame(threadName);
            } else {
              virtualStack.pop();
            }
          }
          this.currentEvent++;
          IStep.StepEvent _get = this.stepEvents.get(this.currentEvent);
          event = _get;
          Map<String, Object> _parameters_1 = event.step.getParameters();
          Object _get_1 = _parameters_1.get("this");
          obj = _get_1;
        }
      }
      boolean _equals = Objects.equal(obj, step);
      if (_equals) {
        final Iterator<IStep> itr = virtualStack.descendingIterator();
        while (itr.hasNext()) {
          IStep _next = itr.next();
          this.pushStackFrame(threadName, _next);
        }
        if (event.start) {
          this.pushStackFrame(threadName, event.step);
        } else {
          this.currentEvent++;
          this.popStackFrame(threadName);
          this.findNextStartedStep(threadName);
        }
      } else {
        throw new IllegalStateException("State events ended before the desired event was found");
      }
    }
  }
  
  @Override
  protected void updateStack(final String threadName, final EObject instruction) {
    if ((!this.inThePast)) {
      super.updateStack(threadName, instruction);
    }
  }
  
  @Override
  public void resume() {
    if ((!this.executionTerminated)) {
      if (this.inThePast) {
        this.loadLastState(this.threadName);
      }
      super.resume();
    }
  }
  
  @Override
  public void resume(final String threadName) {
    if ((!this.executionTerminated)) {
      if (this.inThePast) {
        this.loadLastState(threadName);
      }
      super.resume(threadName);
    }
  }
  
  @Override
  public void terminate() {
    super.terminate();
    Activator _default = Activator.getDefault();
    final Supplier<OmniscientGenericSequentialModelDebugger> _function = new Supplier<OmniscientGenericSequentialModelDebugger>() {
      @Override
      public OmniscientGenericSequentialModelDebugger get() {
        return null;
      }
    };
    _default.setDebuggerSupplier(_function);
  }
  
  @Override
  public void stepInto(final String threadName) {
    if (this.inThePast) {
      final int size = this.stepEvents.size();
      if ((this.currentEvent < (size - 1))) {
        this.currentEvent++;
        IStep.StepEvent event = this.stepEvents.get(this.currentEvent);
        if (event.start) {
          this.pushStackFrame(threadName, event.step);
        } else {
          this.findNextStartedStep(threadName);
        }
      } else {
        if ((this.lastJumpIndex == (-1))) {
          this.inThePast = false;
          super.stepInto(threadName);
        } else {
          this.jumpToState((this.lastJumpIndex + 1));
          this.findNextStartedStep(threadName);
        }
      }
    } else {
      super.stepInto(threadName);
    }
  }
  
  @Override
  protected void setupStepOverPredicateBreak() {
    abstract class __OmniscientGenericSequentialModelDebugger_1 implements BiPredicate<IBasicExecutionEngine, MSEOccurrence> {
      MSEOccurrence steppedOver;
    }
    
    if ((this.steppingOverStackFrameIndex != (-1))) {
      final ISequentialExecutionEngine seqEngine = ((ISequentialExecutionEngine) this.engine);
      final Deque<MSEOccurrence> stack = seqEngine.getCurrentStack();
      int _size = stack.size();
      final int idx = (_size - this.steppingOverStackFrameIndex);
      __OmniscientGenericSequentialModelDebugger_1 ___OmniscientGenericSequentialModelDebugger_1 = new __OmniscientGenericSequentialModelDebugger_1() {
        {
          steppedOver = ((MSEOccurrence[])Conversions.unwrapArray(stack, MSEOccurrence.class))[idx];
        }
        @Override
        public boolean test(final IBasicExecutionEngine t, final MSEOccurrence u) {
          Deque<MSEOccurrence> _currentStack = seqEngine.getCurrentStack();
          boolean _contains = _currentStack.contains(this.steppedOver);
          return (!_contains);
        }
      };
      this.addPredicateBreak(___OmniscientGenericSequentialModelDebugger_1);
      this.steppingOverStackFrameIndex = (-1);
    } else {
      super.setupStepOverPredicateBreak();
    }
  }
  
  @Override
  public void stepOver(final String threadName) {
    if (this.inThePast) {
      IStep.StepEvent _get = this.stepEvents.get(this.currentEvent);
      final IStep steppedOver = _get.step;
      final int endingStateIdx = steppedOver.getEndingIndex();
      if ((endingStateIdx == (-1))) {
        this.steppingOverStackFrameIndex = (this.nbStackFrames - 1);
        this.loadLastState(threadName);
        super.stepOver(threadName);
      } else {
        int _currentStateIndex = this.getCurrentStateIndex();
        boolean _greaterThan = (endingStateIdx > _currentStateIndex);
        if (_greaterThan) {
          this.jumpToState(endingStateIdx);
        } else {
          this.currentEvent++;
        }
        Map<String, Object> _parameters = steppedOver.getParameters();
        Object _get_1 = _parameters.get("this");
        this.findCorrespondingStepEvent(threadName, _get_1);
      }
    } else {
      super.stepOver(threadName);
    }
  }
  
  @Override
  protected void setupStepReturnPredicateBreak() {
    abstract class __OmniscientGenericSequentialModelDebugger_2 implements BiPredicate<IBasicExecutionEngine, MSEOccurrence> {
      MSEOccurrence steppedReturn;
    }
    
    if ((this.steppingReturnStackFrameIndex != (-1))) {
      final ISequentialExecutionEngine seqEngine = ((ISequentialExecutionEngine) this.engine);
      final Deque<MSEOccurrence> stack = seqEngine.getCurrentStack();
      int _size = stack.size();
      final int idx = (_size - this.steppingReturnStackFrameIndex);
      __OmniscientGenericSequentialModelDebugger_2 ___OmniscientGenericSequentialModelDebugger_2 = new __OmniscientGenericSequentialModelDebugger_2() {
        {
          steppedReturn = ((MSEOccurrence[])Conversions.unwrapArray(stack, MSEOccurrence.class))[idx];
        }
        @Override
        public boolean test(final IBasicExecutionEngine t, final MSEOccurrence u) {
          Deque<MSEOccurrence> _currentStack = seqEngine.getCurrentStack();
          boolean _contains = _currentStack.contains(this.steppedReturn);
          return (!_contains);
        }
      };
      this.addPredicateBreak(___OmniscientGenericSequentialModelDebugger_2);
      this.steppingReturnStackFrameIndex = (-1);
    } else {
      super.setupStepReturnPredicateBreak();
    }
  }
  
  @Override
  public void stepReturn(final String threadName) {
    if (this.inThePast) {
      IStep.StepEvent _get = this.stepEvents.get(this.currentEvent);
      final IStep currentStep = _get.step;
      final IStep parentStep = currentStep.getParentStep();
      boolean _notEquals = (!Objects.equal(parentStep, null));
      if (_notEquals) {
        final int endIndex = parentStep.getEndingIndex();
        Map<String, Object> _parameters = parentStep.getParameters();
        final Object steppedReturnObject = _parameters.get("this");
        if ((endIndex == (-1))) {
          this.steppingReturnStackFrameIndex = (this.nbStackFrames - 2);
          this.loadLastState(threadName);
          this.inThePast = false;
          super.stepReturn(threadName);
        } else {
          int _currentStateIndex = this.getCurrentStateIndex();
          boolean _notEquals_1 = (endIndex != _currentStateIndex);
          if (_notEquals_1) {
            this.jumpToState(endIndex);
          }
          this.findCorrespondingStepEvent(threadName, steppedReturnObject);
        }
      }
    } else {
      super.stepReturn(threadName);
    }
  }
  
  public boolean canStepBackInto(final String threadName) {
    boolean _xifexpression = false;
    boolean _or = false;
    if ((this.lastJumpIndex > 0)) {
      _or = true;
    } else {
      boolean _and = false;
      if (!(this.lastJumpIndex == (-1))) {
        _and = false;
      } else {
        int _lastIndex = this.getLastIndex();
        boolean _greaterThan = (_lastIndex > 0);
        _and = _greaterThan;
      }
      _or = _and;
    }
    if (_or) {
      _xifexpression = true;
    } else {
      boolean _xblockexpression = false;
      {
        final ArrayList<IStep.StepEvent> events = new ArrayList<IStep.StepEvent>();
        if (this.inThePast) {
          events.addAll(this.stepEvents);
        } else {
          IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
          int _lastIndex_1 = this.getLastIndex();
          List<IStep.StepEvent> _eventsForState = _traceManager.getEventsForState(_lastIndex_1);
          events.addAll(_eventsForState);
        }
        int _size = events.size();
        int itr = (_size - 1);
        IStep.StepEvent event = events.get(itr);
        while ((event.start && (itr > (-1)))) {
          {
            itr--;
            IStep.StepEvent _get = events.get(itr);
            event = _get;
          }
        }
        _xblockexpression = (!event.start);
      }
      _xifexpression = _xblockexpression;
    }
    return _xifexpression;
  }
  
  public boolean canStepBackOver(final String threadName) {
    final ArrayList<IStep.StepEvent> events = new ArrayList<IStep.StepEvent>();
    if (this.inThePast) {
      events.addAll(this.stepEvents);
    } else {
      IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
      int _lastIndex = this.getLastIndex();
      List<IStep.StepEvent> _eventsForState = _traceManager.getEventsForState(_lastIndex);
      events.addAll(_eventsForState);
    }
    int _size = events.size();
    int itr = (_size - 1);
    IStep.StepEvent event = events.get(itr);
    while ((event.start && (itr > (-1)))) {
      {
        itr--;
        IStep.StepEvent _get = events.get(itr);
        event = _get;
      }
    }
    return (!event.start);
  }
  
  public boolean canStepBackOut(final String threadName) {
    final ArrayList<IStep.StepEvent> events = new ArrayList<IStep.StepEvent>();
    if (this.inThePast) {
      events.addAll(this.stepEvents);
    } else {
      IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
      int _lastIndex = this.getLastIndex();
      List<IStep.StepEvent> _eventsForState = _traceManager.getEventsForState(_lastIndex);
      events.addAll(_eventsForState);
    }
    int _size = events.size();
    int itr = (_size - 1);
    IStep.StepEvent event = events.get(itr);
    while ((event.start && (itr > (-1)))) {
      {
        itr--;
        IStep.StepEvent _get = events.get(itr);
        event = _get;
      }
    }
    boolean _and = false;
    if (!(!event.start)) {
      _and = false;
    } else {
      IStep _parentStep = event.step.getParentStep();
      boolean _notEquals = (!Objects.equal(_parentStep, null));
      _and = _notEquals;
    }
    return _and;
  }
  
  public void stepBackInto(final String threadName) {
    if ((!this.inThePast)) {
      this.inThePast = true;
      int _lastIndex = this.getLastIndex();
      this.updateStateEvents(_lastIndex);
      int _size = this.stepEvents.size();
      int _minus = (_size - 1);
      this.currentEvent = _minus;
    }
    this.findPreviousStartedStep(threadName);
  }
  
  public void stepBackOver(final String threadName) {
    if ((!this.inThePast)) {
      this.inThePast = true;
      int _lastIndex = this.getLastIndex();
      this.updateStateEvents(_lastIndex);
      int _size = this.stepEvents.size();
      int _minus = (_size - 1);
      this.currentEvent = _minus;
    }
    final IStep step = this.findPreviousEndedStep(threadName);
    boolean _notEquals = (!Objects.equal(step, null));
    if (_notEquals) {
      int _startingIndex = step.getStartingIndex();
      this.jumpToState(_startingIndex);
      Map<String, Object> _parameters = step.getParameters();
      Object _get = _parameters.get("this");
      this.findCorrespondingStepEvent(threadName, _get);
    }
  }
  
  public void stepBackOut(final String threadName) {
    if ((!this.inThePast)) {
      this.inThePast = true;
      int _lastIndex = this.getLastIndex();
      this.updateStateEvents(_lastIndex);
      int _size = this.stepEvents.size();
      int _minus = (_size - 1);
      this.currentEvent = _minus;
    }
    final IStep step = this.findPreviousEndedStep(threadName);
    boolean _and = false;
    boolean _notEquals = (!Objects.equal(step, null));
    if (!_notEquals) {
      _and = false;
    } else {
      IStep _parentStep = step.getParentStep();
      boolean _notEquals_1 = (!Objects.equal(_parentStep, null));
      _and = _notEquals_1;
    }
    if (_and) {
      final IStep parentStep = step.getParentStep();
      int _startingIndex = parentStep.getStartingIndex();
      this.jumpToState(_startingIndex);
      Map<String, Object> _parameters = parentStep.getParameters();
      Object _get = _parameters.get("this");
      this.findCorrespondingStepEvent(threadName, _get);
    }
  }
  
  public boolean canStepValue(final int traceIndex) {
    return true;
  }
  
  private int getNextValueIndex(final IValueTrace valueTrace) {
    int stateIndex = this.getCurrentStateIndex();
    final int currentValueIndex = valueTrace.getActiveValueIndex(stateIndex);
    int valueIndex = currentValueIndex;
    while (((stateIndex < this.getLastIndex()) && ((valueIndex == currentValueIndex) || (valueIndex == (-1))))) {
      {
        stateIndex++;
        int _activeValueIndex = valueTrace.getActiveValueIndex(stateIndex);
        valueIndex = _activeValueIndex;
      }
    }
    if (((valueIndex == currentValueIndex) || (valueIndex == (-1)))) {
      return valueTrace.getSize();
    }
    return valueIndex;
  }
  
  public void stepValue(final int traceIndex) {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    List<IValueTrace> _allValueTraces = _traceManager.getAllValueTraces();
    final IValueTrace valueTrace = _allValueTraces.get(traceIndex);
    final int i = this.getNextValueIndex(valueTrace);
    boolean _and = false;
    int _size = valueTrace.getSize();
    boolean _lessThan = (i < _size);
    if (!_lessThan) {
      _and = false;
    } else {
      _and = (i != (-1));
    }
    if (_and) {
      EObject _value = valueTrace.getValue(i);
      this.jump(_value);
    } else {
      this.setupStepValuePredicateBreak(valueTrace, i);
      this.resume();
    }
  }
  
  private void setupStepValuePredicateBreak(final IValueTrace valueTrace, final int valueIndex) {
    this.addPredicateBreak(new BiPredicate<IBasicExecutionEngine, MSEOccurrence>() {
      @Override
      public boolean test(final IBasicExecutionEngine t, final MSEOccurrence u) {
        int _currentStateIndex = OmniscientGenericSequentialModelDebugger.this.getCurrentStateIndex();
        final int i = valueTrace.getActiveValueIndex(_currentStateIndex);
        final int j = valueIndex;
        return (i == j);
      }
    });
  }
  
  public boolean canBackValue(final int traceIndex) {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    final List<IValueTrace> allValueTraces = _traceManager.getAllValueTraces();
    boolean _and = false;
    int _size = allValueTraces.size();
    boolean _lessThan = (traceIndex < _size);
    if (!_lessThan) {
      _and = false;
    } else {
      _and = (traceIndex > (-1));
    }
    if (_and) {
      final IValueTrace valueTrace = allValueTraces.get(traceIndex);
      int _currentStateIndex = this.getCurrentStateIndex();
      final int currentValueIndex = valueTrace.getActiveValueIndex(_currentStateIndex);
      int stateIndex = this.getCurrentStateIndex();
      int valueIndex = valueTrace.getActiveValueIndex(stateIndex);
      while (((stateIndex > 0) && ((valueIndex == currentValueIndex) || (valueIndex == (-1))))) {
        {
          stateIndex--;
          int _activeValueIndex = valueTrace.getActiveValueIndex(stateIndex);
          valueIndex = _activeValueIndex;
        }
      }
      return ((valueIndex != currentValueIndex) && (valueIndex != (-1)));
    }
    return false;
  }
  
  private int getPreviousValueIndex(final IValueTrace valueTrace) {
    int _currentStateIndex = this.getCurrentStateIndex();
    final int currentValueIndex = valueTrace.getActiveValueIndex(_currentStateIndex);
    int _currentStateIndex_1 = this.getCurrentStateIndex();
    int stateIndex = (_currentStateIndex_1 - 1);
    int valueIndex = valueTrace.getActiveValueIndex(stateIndex);
    while (((stateIndex > 0) && ((valueIndex == currentValueIndex) || (valueIndex == (-1))))) {
      {
        stateIndex--;
        int _activeValueIndex = valueTrace.getActiveValueIndex(stateIndex);
        valueIndex = _activeValueIndex;
      }
    }
    return valueIndex;
  }
  
  public void backValue(final int traceIndex) {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    List<IValueTrace> _allValueTraces = _traceManager.getAllValueTraces();
    final IValueTrace valueTrace = _allValueTraces.get(traceIndex);
    int _previousValueIndex = this.getPreviousValueIndex(valueTrace);
    EObject _value = valueTrace.getValue(_previousValueIndex);
    this.jump(_value);
  }
  
  /**
   * Returns the index of the last state in the trace.
   */
  private int getLastIndex() {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    int _traceSize = _traceManager.getTraceSize();
    return (_traceSize - 1);
  }
  
  /**
   * If we are in the past, we read the last jump index.
   * Otherwise, we read the trace size.
   */
  public int getCurrentStateIndex() {
    if ((this.lastJumpIndex != (-1))) {
      return this.lastJumpIndex;
    } else {
      return this.getLastIndex();
    }
  }
  
  private void jumpToState(final int i) {
    while ((this.nbStackFrames > 1)) {
      this.popStackFrame(this.threadName);
    }
    this.setCurrentInstruction(this.threadName, this.executedModelRoot);
    this.inThePast = true;
    int _lastIndex = this.getLastIndex();
    boolean _equals = (i == _lastIndex);
    if (_equals) {
      this.lastJumpIndex = (-1);
    } else {
      this.lastJumpIndex = i;
    }
    this.traceAddon.goTo(i);
    this.updateStateEvents(i);
    this.updateData(this.threadName, null);
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    int _currentStateIndex = this.getCurrentStateIndex();
    final List<IStep> beforeStack = _traceManager.getStackForwardBeforeState(_currentStateIndex);
    final Consumer<IStep> _function = new Consumer<IStep>() {
      @Override
      public void accept(final IStep s) {
        OmniscientGenericSequentialModelDebugger.this.pushStackFrame(OmniscientGenericSequentialModelDebugger.this.threadName, s);
      }
    };
    beforeStack.forEach(_function);
  }
  
  /**
   * To be used by the timeline
   */
  public void jump(final int i) {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    int _traceSize = _traceManager.getTraceSize();
    boolean _lessThan = (i < _traceSize);
    if (_lessThan) {
      this.jumpToState(i);
      this.findNextStartedStep(this.threadName);
    }
    this.scheduleSelectLastStackframe(500);
  }
  
  /**
   * To be used by the timeline
   */
  public void jump(final EObject o) {
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    int _stateOrValueIndex = _traceManager.getStateOrValueIndex(o);
    this.jump(_stateOrValueIndex);
  }
  
  @Override
  public boolean validateVariableValue(final String threadName, final String variableName, final String value) {
    if (this.inThePast) {
      Status _status = new Status(IStatus.ERROR, Activator.PLUGIN_ID, "Illegal variable value set");
      ErrorDialog.openError(null, "Illegal variable value set", "Cannot set the value of a variable when in replay mode", _status);
      return false;
    }
    return super.validateVariableValue(threadName, variableName, value);
  }
  
  @Override
  public void engineStarted(final IBasicExecutionEngine executionEngine) {
    final Activator activator = Activator.getDefault();
    final Supplier<OmniscientGenericSequentialModelDebugger> _function = new Supplier<OmniscientGenericSequentialModelDebugger>() {
      @Override
      public OmniscientGenericSequentialModelDebugger get() {
        return OmniscientGenericSequentialModelDebugger.this;
      }
    };
    activator.setDebuggerSupplier(_function);
    super.engineStarted(executionEngine);
  }
  
  @Override
  public void engineAboutToStop(final IBasicExecutionEngine engine) {
    super.engineAboutToStop(engine);
    IGemocTraceManager _traceManager = this.traceAddon.getTraceManager();
    int _traceSize = _traceManager.getTraceSize();
    int _minus = (_traceSize - 1);
    this.jumpToState(_minus);
    int _size = this.stepEvents.size();
    int _minus_1 = (_size - 1);
    this.currentEvent = _minus_1;
  }
  
  @Override
  public void engineStopped(final IBasicExecutionEngine executionEngine) {
    final Activator activator = Activator.getDefault();
    activator.setDebuggerSupplier(null);
    super.engineStopped(executionEngine);
  }
}
